/*
 * linux/arch/arm/drivers/scsi/fas216.c
 *
 * Copyright (C) 1997 Russell King
 *
 * Based in information in qlogicfas.c by Tom Zerucha, Michael Griffith, and
 * other sources.
 *
 * This is a generic driver.  To use it, have a look at cumana_2.c.  You
 * should define your own structure that overlays FAS216_Info, eg:
 * struct my_host_data {
 *    FAS216_Info info;
 *    ... my host specific data ...
 * };
 *
 * Changelog:
 *  30-08-1997	RMK	Created
 *  14-09-1997	RMK	Started disconnect support
 */

#define __NO_VERSION__
 
#include <linux/module.h>
#include <linux/blk.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/unistd.h>
#include <linux/stat.h>

#include <asm/delay.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/ecard.h>

#define FAS216_C

#include "scsi.h"
#include "hosts.h"
#include "fas216.h"

#define VER_MAJOR	0
#define VER_MINOR	0
#define VER_PATCH	1

#undef NO_DISCONNECTS
#define DEBUG_CONNECT
#undef DEBUG_BUSSERVICE
#undef DEBUG_FUNCTIONDONE
#define DEBUG_MESSAGES

static char *fas216_bus_phase (int stat)
{
    static char *phases[] = {
	"DATA OUT", "DATA IN",
	"COMMAND", "STATUS",
	"MISC OUT", "MISC IN",
	"MESG OUT", "MESG IN"
    };

    return phases[stat & STATUS_BUSMASK];
}

static char fas216_target (FAS216_Info *info)
{
    if (info->SCpnt)
	return '0' + info->SCpnt->target;
    else
	return 'H';
}

static void fas216_done (FAS216_Info *info, unsigned int result);

/*
 * Function: void fas216_updateptrs (FAS216_Info *info, int bytes_transferred)
 * Purpose : update data pointers after transfer suspended/paused
 * Params  : info              - interface's local pointer to update
 *           bytes_transferred - number of bytes transferred
 */
static void
fas216_updateptrs (FAS216_Info *info, int bytes_transferred)
{
    unsigned char *ptr = info->scsi.SCp.ptr;
    unsigned int residual = info->scsi.SCp.this_residual;

    residual -= bytes_transferred;

    if (residual == 0) {
	if (info->scsi.SCp.buffers_residual) {
	    info->scsi.SCp.buffer++;
	    info->scsi.SCp.buffers_residual--;
	    ptr = (unsigned char *)info->scsi.SCp.buffer->address;
	    residual = info->scsi.SCp.buffer->length;
	} else {
	    ptr = NULL;
	    residual = 0;
	}
    } else
	ptr += bytes_transferred;

    info->scsi.SCp.ptr = ptr;
    info->scsi.SCp.this_residual = residual;
}

/*
 * Function: void fas216_pio (FAS216_Info *info, dmadir_t direction)
 * Purpose : transfer data off of/on to card using programmed IO
 * Params  : info      - interface to transfer data to/from
 *           direction - direction to transfer data (DMA_OUT/DMA_IN)
 * Notes   : this is incredibly slow
 */
static void
fas216_pio (FAS216_Info *info, dmadir_t direction)
{
    unsigned int length = info->scsi.SCp.this_residual;
    char *ptr = info->scsi.SCp.ptr;

    if (direction == DMA_OUT) {
    	while (length > 0) {
	    if ((inb (info->scsi.io_port + REG0_FIFOSTATE) & FF_BYTES) != 0) {
		outb (*ptr++, info->scsi.io_port + REG0_FIFODATA);
		length -= 1;
	    } else if (inb (info->scsi.io_port + REG0_STATUS) & STATUS_INT)
		break;
	}
    } else {
    	while (length > 0) {
	    if ((inb (info->scsi.io_port + REG0_FIFOSTATE) & FF_BYTES) != 0) {
		*ptr++ = inb (info->scsi.io_port + REG0_FIFODATA);
		length -= 1;
	    } else if (inb (info->scsi.io_port + REG0_STATUS) & STATUS_INT)
		break;
	}
    }

    if (length == 0) {
	if (info->scsi.SCp.buffers_residual) {
	    info->scsi.SCp.buffer++;
	    info->scsi.SCp.buffers_residual--;
	    ptr = (unsigned char *)info->scsi.SCp.buffer->address;
	    length = info->scsi.SCp.buffer->length;
	} else {
	    ptr = NULL;
	    length = 0;
	}
    }

    info->scsi.SCp.ptr = ptr;
    info->scsi.SCp.this_residual = length;
}

/*
 * Function: void fas216_starttransfer (FAS216_Info *info, dmadir_t direction)
 * Purpose : Start a DMA/PIO transfer off of/on to card
 * Params  : info      - interface from which device disconnected from
 *           direction - transfer direction (DMA_OUT/DMA_IN)
 */
static void
fas216_starttransfer (FAS216_Info *info, dmadir_t direction)
{
    if (direction == DMA_OUT)
	info->scsi.phase = PHASE_DATAOUT;
    else
	info->scsi.phase = PHASE_DATAIN;

    if (!info->scsi.SCp.ptr /*&& info->scsi.SCp.this_residual*/) {
	printk ("scsi%d.%c: null buffer passed to fas216_starttransfer\n",
		info->host->host_no, fas216_target (info));
	return;
    }

    outb (info->scsi.SCp.this_residual, info->scsi.io_port + REG0_LOCOUNT);
    outb (info->scsi.SCp.this_residual >> 8, info->scsi.io_port + REG0_MDCOUNT);
    outb (info->scsi.SCp.this_residual >> 16, info->scsi.io_port + REG0_HICOUNT);
    outb (CMD_NOP | CMD_WITHDMA, info->scsi.io_port + REG0_CMD);

    if (info->dma.setup) {
	if (info->dma.setup (info->host, &info->scsi.SCp, direction))
	    outb (CMD_TRANSFERINFO | CMD_WITHDMA, info->scsi.io_port + REG0_CMD);
	else {
	    int transferred;

	    outb (CMD_TRANSFERINFO | CMD_WITHDMA, info->scsi.io_port + REG0_CMD);
	    transferred = info->dma.pseudo (info->host, &info->scsi.SCp, direction,
				info->SCpnt->transfersize);

	    fas216_updateptrs (info, transferred);
	}
    } else {
	outb (CMD_TRANSFERINFO, info->scsi.io_port + REG0_CMD);
	fas216_pio (info, direction);
    }
}

/*
 * Function: void fas216_disconnected_intr (FAS216_Info *info)
 * Purpose : handle device disconnection
 * Params  : info - interface from which device disconnected from
 */
static void
fas216_disconnect_intr (FAS216_Info *info)
{
#ifdef DEBUG_CONNECT
    printk ("scsi%d.%c: disconnect\n", info->host->host_no, fas216_target (info));
#endif
    msgqueue_flush (&info->scsi.msgs);

    switch (info->scsi.phase) {
    case PHASE_SELECTION:			/* while selecting - no target		*/
	fas216_done (info, DID_NO_CONNECT);
	break;

    case PHASE_DISCONNECT:			/* message in - disconnecting		*/
	outb (CMD_ENABLESEL, info->scsi.io_port + REG0_CMD);
	info->scsi.disconnectable = 1;
	info->scsi.reconnected.tag = 0;
	info->scsi.phase = PHASE_IDLE;
	info->stats.disconnects += 1;
	break;

    case PHASE_DONE:				/* at end of command - complete		*/
	fas216_done (info, DID_OK);
	break;

    case PHASE_AFTERMSGOUT:			/* message out - possible ABORT message	*/
	if (info->scsi.last_message == ABORT) {
	    info->scsi.aborting = 0;
	    fas216_done (info, DID_ABORT);
	    break;
	}

    default:					/* huh?					*/
	printk (KERN_ERR "scsi%d.%c: unexpected disconnect\n",
		info->host->host_no, fas216_target (info));
	fas216_done (info, DID_ERROR);
	break;
    }
}

/*
 * Function: void fas216_reselected_intr (FAS216_Info *info)
 * Purpose : Start reconnection of a device
 * Params  : info - interface which was reselected
 */
static void
fas216_reselected_intr (FAS216_Info *info)
{
    unsigned char target, identify_msg, ok;

    msgqueue_flush (&info->scsi.msgs);

    if ((inb (info->scsi.io_port + REG0_FIFOSTATE) & FF_BYTES) != 2) {
	printk (KERN_ERR "scsi%d.H: incorrect number of bytes after reselect\n",
		info->host->host_no);
	outb (CMD_SETATN, info->scsi.io_port + REG0_CMD);
	msgqueue_addmsg (&info->scsi.msgs, 1, MESSAGE_REJECT);
	info->scsi.phase = PHASE_MSGOUT;
	outb (CMD_MSGACCEPTED, info->scsi.io_port + REG0_CMD);
	return;
    }

    target = inb (info->scsi.io_port + REG0_FIFODATA);
    identify_msg = inb (info->scsi.io_port + REG0_FIFODATA);

    ok = 1;
    if (!(target & (1 << info->host->this_id))) {
	printk (KERN_ERR "scsi%d.H: invalid host id on reselect\n", info->host->host_no);
	ok = 0;
    }

    if (!(identify_msg & 0x80)) {
	printk (KERN_ERR "scsi%d.H: no IDENTIFY message on reselect, got msg %02X\n",
		info->host->host_no, identify_msg);
	ok = 0;
    }

    if (!ok) {
	/*
	 * Something went wrong - abort the command on
	 * the target.  Should this be INITIATOR_ERROR ?
	 */
	outb (CMD_SETATN, info->scsi.io_port + REG0_CMD);
	msgqueue_addmsg (&info->scsi.msgs, 1, ABORT);
	info->scsi.phase = PHASE_MSGOUT;
	outb (CMD_MSGACCEPTED, info->scsi.io_port + REG0_CMD);
	return;
    }

    target &= ~(1 << info->host->this_id);
    switch (target) {
    case   1:  target = 0; break;
    case   2:  target = 1; break;
    case   4:  target = 2; break;
    case   8:  target = 3; break;
    case  16:  target = 4; break;
    case  32:  target = 5; break;
    case  64:  target = 6; break;
    case 128:  target = 7; break;
    default:   target = info->host->this_id; break;
    }

    identify_msg &= 7;
    info->scsi.reconnected.target = target;
    info->scsi.reconnected.lun    = identify_msg;
    info->scsi.reconnected.tag    = 0;

    ok = 0;
    if (info->scsi.disconnectable && info->SCpnt &&
	  info->SCpnt->target == target && info->SCpnt->lun == identify_msg)
	ok = 1;

    if (!ok && queue_probetgtlun (&info->queues.disconnected, target, identify_msg))
	ok = 1;

    if (ok) {
	info->scsi.phase = PHASE_RECONNECTED;
	outb (target, info->scsi.io_port + REG0_BUSID);
    } else {
	/*
	 * Our command structure not found - abort the command on the target
	 * Should this be INITIATOR_ERROR ?
	 */
	outb (CMD_SETATN, info->scsi.io_port + REG0_CMD);
	msgqueue_addmsg (&info->scsi.msgs, 1, ABORT);
	info->scsi.phase = PHASE_MSGOUT;
    }
    outb (CMD_MSGACCEPTED, info->scsi.io_port + REG0_CMD);
}

/*
 * Function: void fas216_finish_reconnect (FAS216_Info *info)
 * Purpose : finish reconnection sequence for device
 * Params  : info - interface which caused function done interrupt
 */
static void
fas216_finish_reconnect (FAS216_Info *info)
{
#ifdef DEBUG_CONNECT
printk ("Connected: %1X %1X %02X, reconnected: %1X %1X %02X\n",
	info->SCpnt->target, info->SCpnt->lun, info->SCpnt->tag,
	info->scsi.reconnected.target, info->scsi.reconnected.lun,
	info->scsi.reconnected.tag);
#endif

    if (info->scsi.disconnectable && info->SCpnt) {
	info->scsi.disconnectable = 0;
	if (info->SCpnt->target == info->scsi.reconnected.target &&
	    info->SCpnt->lun    == info->scsi.reconnected.lun &&
	    info->SCpnt->tag    == info->scsi.reconnected.tag) {
#ifdef DEBUG_CONNECT
	    printk ("scsi%d.%c: reconnected",
		    info->host->host_no, fas216_target (info));
#endif
	} else {
	    queue_add_cmd_tail (&info->queues.disconnected, info->SCpnt);
#ifdef DEBUG_CONNECT
	    printk ("scsi%d.%c: had to move command to disconnected queue\n",
			info->host->host_no, fas216_target (info));
#endif
	    info->SCpnt = NULL;
	}
    }
    if (!info->SCpnt) {
	info->SCpnt = queue_remove_tgtluntag (&info->queues.disconnected,
				info->scsi.reconnected.target,
				info->scsi.reconnected.lun,
				info->scsi.reconnected.tag);
#ifdef DEBUG_CONNECT
	printk ("scsi%d.%c: had to get command",
		info->host->host_no, fas216_target (info));
#endif
    }
    if (!info->SCpnt) {
	outb (CMD_SETATN, info->scsi.io_port + REG0_CMD);
	msgqueue_addmsg (&info->scsi.msgs, 1, ABORT);
	info->scsi.phase = PHASE_MSGOUT;
	info->scsi.aborting = 1;
    } else {
	/*
	 * Restore data pointer from SAVED data pointer
	 */
	info->scsi.SCp = info->SCpnt->SCp;
#ifdef DEBUG_CONNECT
	printk (", data pointers: [%p, %X]",
		info->scsi.SCp.ptr, info->scsi.SCp.this_residual);
#endif
    }
#ifdef DEBUG_CONNECT
    printk ("\n");
#endif
}

/*
 * Function: void fas216_message (FAS216_Info *info)
 * Purpose : handle a function done interrupt from FAS216 chip
 * Params  : info - interface which caused function done interrupt
 */
static void fas216_message (FAS216_Info *info)
{
    unsigned char message[16];
    unsigned int msglen = 1;

    message[0] = inb (info->scsi.io_port + REG0_FIFODATA);

    if (message[0] == EXTENDED_MESSAGE) {
	message[1] = inb (info->scsi.io_port + REG0_FIFODATA);

	for (msglen = 2; msglen < message[1]; msglen++)
	    message[msglen] = inb (info->scsi.io_port + REG0_FIFODATA);
    }

#ifdef DEBUG_MESSAGES
    {
	int i;

	printk ("scsi%d.%c: message in: ",
		info->host->host_no, fas216_target (info));
	for (i = 0; i < msglen; i++)
	    printk ("%02X ", message[i]);
	printk ("\n");
    }
#endif
    if (info->scsi.phase == PHASE_RECONNECTED) {
	if (message[0] == SIMPLE_QUEUE_TAG)
	    info->scsi.reconnected.tag = message[1];
	fas216_finish_reconnect (info);
	info->scsi.phase = PHASE_MSGIN;
    }	

    switch (message[0]) {
    case COMMAND_COMPLETE:
	printk ("fas216: command complete in MESSAGE_IN?\n");
	break;

    case SAVE_POINTERS:
	/*
	 * Save current data pointer to SAVED data pointer
	 */
	info->SCpnt->SCp = info->scsi.SCp;
#if defined (DEBUG_MESSAGES) || defined (DEBUG_CONNECT)
	printk ("scsi%d.%c: save data pointers: [%p, %X]\n",
		info->host->host_no, fas216_target (info),
		info->scsi.SCp.ptr, info->scsi.SCp.this_residual);
#endif
	break;

    case RESTORE_POINTERS:
	/*
	 * Restore current data pointer from SAVED data pointer
	 */
	info->scsi.SCp = info->SCpnt->SCp;
#if defined (DEBUG_MESSAGES) || defined (DEBUG_CONNECT)
	printk ("scsi%d.%c: restore data pointers: [%p, %X]\n",
		info->host->host_no, fas216_target (info),
		info->scsi.SCp.ptr, info->scsi.SCp.this_residual);
#endif
	break;

    case DISCONNECT:
	info->scsi.phase = PHASE_DISCONNECT;
	break;

    case MESSAGE_REJECT:
	printk ("scsi%d.%c: reject, last message %04X\n",
		info->host->host_no, fas216_target (info),
		info->scsi.last_message);
	break;

    case SIMPLE_QUEUE_TAG:
	printk ("scsi%d.%c: reconnect queue tag %02X\n",
		info->host->host_no, fas216_target (info),
		message[1]);
	break;

    default:
	printk ("scsi%d.%c: unrecognised message %02X, rejecting\n",
		info->host->host_no, fas216_target (info),
		message[0]);
	msgqueue_flush (&info->scsi.msgs);
	outb (CMD_SETATN, info->scsi.io_port + REG0_CMD);
	msgqueue_addmsg (&info->scsi.msgs, 1, MESSAGE_REJECT);
	info->scsi.phase = PHASE_MSGOUT;
	break;
    }
    outb (CMD_MSGACCEPTED, info->scsi.io_port + REG0_CMD);
}

/*
 * Function: void fas216_busservice_intr (FAS216_Info *info, unsigned int stat, unsigned int ssr)
 * Purpose : handle a bus service interrupt from FAS216 chip
 * Params  : info - interface which caused bus service interrupt
 *           stat - Status register contents
 *           ssr  - SCSI Status register contents
 */
static void fas216_busservice_intr (FAS216_Info *info, unsigned int stat, unsigned int ssr)
{
    int i;
#ifdef DEBUG_BUSSERVICE
    printk ("scsi%d.%c: bus service: stat=%02X ssr=%02X\n",
	    info->host->host_no, fas215_target (info), stat, ssr);
#endif
    switch (ssr & SSR_BITS) {
    case SSR_COMPLETE:			/* last action completed		*/
	switch (info->scsi.phase) {
	case PHASE_SELECTION:		/* while selecting - selected target	*/
	    outb (CMD_NOP, info->scsi.io_port + REG0_CMD);
	    switch (stat & STATUS_BUSMASK) {
	    case STATUS_DATAOUT:	/* data out phase			*/
		fas216_starttransfer (info, DMA_OUT);
		break;

	    case STATUS_DATAIN:		/* data in phase			*/
		fas216_starttransfer (info, DMA_IN);
		break;

	    case STATUS_STATUS:		/* status phase				*/
		info->scsi.phase = PHASE_STATUS;
		outb (CMD_INITCMDCOMPLETE, info->scsi.io_port + REG0_CMD);
		break;

	    case STATUS_MESGIN:		/* message in phase			*/
		info->scsi.phase = PHASE_MSGIN;
		outb (CMD_TRANSFERINFO, info->scsi.io_port + REG0_CMD);
		break;

	    default:			/* other				*/
		printk ("scsi%d.%c: bus phase %s after connect?\n",
			info->host->host_no, fas216_target (info),
			fas216_bus_phase (stat));
		break;
	    }
	    break;

	case PHASE_DATAIN:		/* while transfering data in		*/
	    switch (stat & STATUS_BUSMASK) {
	    case STATUS_DATAIN:		/* continue data in phase		*/
		fas216_starttransfer (info, DMA_IN);
		break;

	    case STATUS_STATUS:
		if (info->dma.stop)
		    info->dma.stop (info->host, &info->scsi.SCp);

		info->scsi.phase = PHASE_STATUS;
		outb (CMD_INITCMDCOMPLETE, info->scsi.io_port + REG0_CMD);
		break;

	    case STATUS_MESGIN:		/* message in phase			*/
		if (info->dma.stop)
		    info->dma.stop (info->host, &info->scsi.SCp);

		info->scsi.phase = PHASE_MSGIN;
		outb (CMD_TRANSFERINFO, info->scsi.io_port + REG0_CMD);
		break;

	    default:
		printk ("scsi%d.%c: bus phase %s after data in?\n",
			info->host->host_no, fas216_target (info),
			fas216_bus_phase (stat));
	    }
	    break;

	case PHASE_DATAOUT:		/* while transfering data out		*/
	    switch (stat & STATUS_BUSMASK) {
	    case STATUS_DATAOUT:
		fas216_starttransfer (info, DMA_OUT);
		break;

	    case STATUS_STATUS:
		if (info->dma.stop)
		    info->dma.stop (info->host, &info->scsi.SCp);

		info->scsi.phase = PHASE_STATUS;
		outb (CMD_INITCMDCOMPLETE, info->scsi.io_port + REG0_CMD);
		break;

	    case STATUS_MESGIN:		/* message in phase			*/
		if (info->dma.stop)
		    info->dma.stop (info->host, &info->scsi.SCp);

		/*
		residual = (inb (info->scsi.io_port + REG0_FIFO) & FIFO_CNT) +
			    inb (info->scsi.io_port + REG0_CNTL) +
			    inb (info->scsi.io_port + REG0_CNTM) << 8 +
			    inb (info->scsi.io_port + REG0_CNTH) << 16;
		 */
		info->scsi.phase = PHASE_MSGIN;

		outb (CMD_TRANSFERINFO, info->scsi.io_port + REG0_CMD);
		break;

	    default:
		printk ("scsi%d.%c: bus phase %s after data out?\n",
			info->host->host_no, fas216_target (info),
			fas216_bus_phase (stat));
	    }
	    break;

	case PHASE_RECONNECTED:
	    /*
	     * Command reconnected - if MESGIN, get message - it may be
	     * the tag.  If not, get command out of the disconnected queue
	     */
	    switch (stat & STATUS_BUSMASK) {
	    case STATUS_MESGIN:
		outb (CMD_TRANSFERINFO, info->scsi.io_port + REG0_CMD);
		break;

	    case STATUS_STATUS:
		fas216_finish_reconnect (info);
		info->scsi.phase = PHASE_STATUS;
		outb (CMD_INITCMDCOMPLETE, info->scsi.io_port + REG0_CMD);
		break;

	    case STATUS_DATAOUT:		/* data out phase		*/
		fas216_finish_reconnect (info);
		fas216_starttransfer (info, DMA_OUT);
		break;

	    case STATUS_DATAIN:			/* data in phase		*/
		fas216_finish_reconnect (info);
		fas216_starttransfer (info, DMA_IN);
		break;

	    default:
		printk ("scsi%d.%c: bus phase %s after reconnect?\n",
			info->host->host_no, fas216_target (info),
			fas216_bus_phase (stat));
	    }
	    break;

	case PHASE_MSGIN:
	    switch (stat & STATUS_BUSMASK) {
	    case STATUS_MESGIN:
		outb (CMD_TRANSFERINFO, info->scsi.io_port + REG0_CMD);
		break;

	    default:
		printk ("scsi%d.%c: bus phase %s after message in?\n",
			info->host->host_no, fas216_target (info),
			fas216_bus_phase (stat));
	    }
	    break;

	case PHASE_MSGOUT:
	    if ((stat & STATUS_BUSMASK) != STATUS_MESGOUT) {
		printk ("scsi%d.%c: didn't manage MESSAGE OUT phase\n",
			info->host->host_no, fas216_target (info));
	    } else {
		unsigned int msglen;

		msglen = msgqueue_msglength (&info->scsi.msgs);

		outb (CMD_FLUSHFIFO, info->scsi.io_port + REG0_CMD);

		if (msglen == 0)
		    outb (NOP, info->scsi.io_port + REG0_FIFODATA);
		else {
		    char *msg;

		    while ((msg = msgqueue_getnextmsg (&info->scsi.msgs, &msglen)) != NULL) {
			for (i = 0; i < msglen; i++)
			    outb (msg[i], info->scsi.io_port + REG0_FIFODATA);
		    }
		}
		outb (CMD_TRANSFERINFO, info->scsi.io_port + REG0_CMD);
		info->scsi.phase = PHASE_AFTERMSGOUT;
	    }
	    break;

	case PHASE_AFTERMSGOUT:
	    switch (stat & STATUS_BUSMASK) {
	    case STATUS_MESGIN:
		info->scsi.phase = PHASE_MSGIN;

		outb (CMD_TRANSFERINFO, info->scsi.io_port + REG0_CMD);
		break;

	    default:
		printk ("scsi%d.%c: bus phase %s after message out\n",
			info->host->host_no, fas216_target (info),
			fas216_bus_phase (stat));
	    }
	    break;

	case PHASE_DISCONNECT:
	    printk ("scsi%d.%c: disconnect message received, but bus service %s?\n",
		    info->host->host_no, fas216_target (info),
		    fas216_bus_phase (stat));
	    outb (CMD_SETATN, info->scsi.io_port + REG0_CMD);
	    msgqueue_addmsg (&info->scsi.msgs, 1, ABORT);
	    info->scsi.phase = PHASE_MSGOUT;
	    info->scsi.aborting = 1;
	    outb (CMD_TRANSFERINFO, info->scsi.io_port + REG0_CMD);
	    break;

	default:
	    printk ("scsi%d.%c: internal phase %d for bus service?"
		    "  What do I do with this?\n",
		    info->host->host_no, fas216_target (info),
		    info->scsi.phase);
	}
	break;

    default:
	printk ("scsi%d.%c: bus service at step %d?\n",
		info->host->host_no, fas216_target (info),
		ssr & SSR_BITS);
    }
}

/*
 * Function: void fas216_funcdone_intr (FAS216_Info *info, unsigned int stat, unsigned int ssr)
 * Purpose : handle a function done interrupt from FAS216 chip
 * Params  : info - interface which caused function done interrupt
 *           stat - Status register contents
 *           ssr  - SCSI Status register contents
 */
static void fas216_funcdone_intr (FAS216_Info *info, unsigned int stat, unsigned int ssr)
{
    int status, message;
#ifdef DEBUG_FUNCTIONDONE
    printk ("scsi%d.%c: function done: stat=%X, ssr=%X\n",
	    info->host->host_no, fas216_target (info), stat, ssr);
#endif
    switch (info->scsi.phase) {
    case PHASE_STATUS:			/* status phase - read status and msg	*/
	status = inb (info->scsi.io_port + REG0_FIFODATA);
	message = inb (info->scsi.io_port + REG0_FIFODATA);
	info->scsi.SCp.Message = message;
	info->scsi.SCp.Status = status;
	info->scsi.phase = PHASE_DONE;
	outb (CMD_MSGACCEPTED, info->scsi.io_port + REG0_CMD);
	break;

    case PHASE_IDLE:			/* reselected?				*/
    case PHASE_MSGIN:			/* message in phase			*/
    case PHASE_RECONNECTED:		/* reconnected command			*/
	if ((stat & STATUS_BUSMASK) == STATUS_MESGIN) {
	    fas216_message (info);
	    break;
	}

    default:
	printk ("scsi%d.%c: internal phase %d for function done?"
		"  What do I do with this?\n",
		info->host->host_no, fas216_target (info),
		info->scsi.phase);
    }
}

/*
 * Function: void fas216_intr (struct Scsi_Host *instance)
 * Purpose : handle interrupts from the interface to progress a command
 * Params  : instance - interface to service
 */
void fas216_intr (struct Scsi_Host *instance)
{
    FAS216_Info *info = (FAS216_Info *)instance->hostdata;
    unsigned char isr, ssr, stat;

    stat = inb (info->scsi.io_port + REG0_STATUS);
    ssr = inb (info->scsi.io_port + REG0_SSR);
    isr = inb (info->scsi.io_port + REG0_ISR);

    if (isr & ISR_BUSRESET)
	printk ("scsi%d.H: fas216: bus reset detected\n", instance->host_no);
    else if (isr & ISR_ILLEGALCMD)
	printk (KERN_CRIT "scsi%d.H: illegal command given\n", instance->host_no);
    else if (isr & ISR_DISCONNECT)
	fas216_disconnect_intr (info);
    else if (isr & ISR_RESELECTED)		/* reselected			*/
	fas216_reselected_intr (info);
    else if (isr & ISR_BUSSERVICE)		/* bus service request		*/
	fas216_busservice_intr (info, stat, ssr);
    else if (isr & ISR_FUNCDONE)		/* function done		*/
	fas216_funcdone_intr (info, stat, ssr);
    else
    	printk ("scsi%d.%c: unknown interrupt received:"
		" phase %d isr %02X ssr %02X stat %02X\n",
		instance->host_no, fas216_target (info),
		info->scsi.phase, isr, ssr, stat);
}

/*
 * Function: void fas216_kick (FAS216_Info *info)
 * Purpose : kick a command to the interface - interface should be idle
 * Params  : info - our host interface to kick
 * Notes   : Interrupts are always disabled!
 */
static void fas216_kick (FAS216_Info *info)
{
    Scsi_Cmnd *SCpnt;
    int i, msglen, from_queue = 0;

    if (info->origSCpnt) {
	SCpnt = info->origSCpnt;
	info->origSCpnt = NULL;
    } else
	SCpnt = NULL;

    /* retrieve next command */
    if (!SCpnt) {
	SCpnt = queue_remove_exclude (&info->queues.issue, info->busyluns);
	from_queue = 1;
    }

    if (!SCpnt) /* no command pending - just exit */
	return;

    if (info->scsi.disconnectable && info->SCpnt) {
	queue_add_cmd_tail (&info->queues.disconnected, info->SCpnt);
	info->scsi.disconnectable = 0;
	info->SCpnt = NULL;
	printk ("scsi%d.%c: moved command to disconnected queue\n",
		info->host->host_no, fas216_target (info));
    }

    /*
     * tagged queuing - allocate a new tag to this command
     */
    if (SCpnt->device->tagged_queue && SCpnt->cmnd[0] != REQUEST_SENSE) {
	SCpnt->device->current_tag += 1;
	if (SCpnt->device->current_tag == 0)
	    SCpnt->device->current_tag = 1;
	SCpnt->tag = SCpnt->device->current_tag;
    }

    /*
     * claim host busy
     */
    info->scsi.phase = PHASE_SELECTION;
    info->SCpnt = SCpnt;
    info->scsi.SCp = SCpnt->SCp;

#ifdef DEBUG_CONNECT
    printk ("scsi%d.%c: starting cmd %02X",
	    info->host->host_no, '0' + SCpnt->target,
	    SCpnt->cmnd[0]);
#endif

    if (from_queue) {
#ifdef SCSI2_TAG
	if (SCpnt->device->tagged_queue && SCpnt->cmnd[0] != REQUEST_SENSE) {
	    SCpnt->device->current_tag += 1;
	    if (SCpnt->device->current_tag == 0)
		SCpnt->device->current_tag = 1;
	    SCpnt->tag = SCpnt->device->current_tag;
	} else
#endif
	    set_bit (SCpnt->target * 8 + SCpnt->lun, info->busyluns);

	info->stats.removes += 1;
    }

    /*
     * build outgoing message bytes
     */
    msgqueue_flush (&info->scsi.msgs);
#ifndef NO_DISCONNECTS
    msgqueue_addmsg (&info->scsi.msgs, 1, IDENTIFY(1, SCpnt->lun));
#else
    msgqueue_addmsg (&info->scsi.msgs, 1, IDENTIFY(0, SCpnt->lun));
#endif

    /*
     * Add tag message if required
     */
    if (SCpnt->tag)
	msgqueue_addmsg (&info->scsi.msgs, 2, SIMPLE_QUEUE_TAG, SCpnt->tag);

    /*
     * following what the ESP driver says...
     */
    outb (0, info->scsi.io_port + REG0_LOCOUNT);
    outb (0, info->scsi.io_port + REG0_MDCOUNT);
    outb (0, info->scsi.io_port + REG0_HICOUNT);
    outb (CMD_NOP | CMD_WITHDMA, info->scsi.io_port + REG0_CMD);

    /*
     * flush FIFO
     */
    outb (CMD_FLUSHFIFO, info->scsi.io_port + REG0_CMD);

    /*
     * load bus-id and timeout
     */
    outb (BUSID(SCpnt->target), info->scsi.io_port + REG0_BUSID);
    outb (info->ifcfg.select_timeout, info->scsi.io_port + REG0_TIMEOUT);

    /*
     * synchronous transfers?
     */
    outb (0, info->scsi.io_port + REG0_SYNCOFFSET);
    outb (info->ifcfg.clockrate / 5, info->scsi.io_port + REG0_PERIOD);

    msglen = msgqueue_msglength (&info->scsi.msgs);

    if (msglen == 1 || msglen == 3) {
	/*
	 * load message bytes
	 */
	char *msg;

	while ((msg = msgqueue_getnextmsg (&info->scsi.msgs, &msglen)) != NULL) {
	    for (i = 0; i < msglen; i++)
		outb (msg[i], info->scsi.io_port + REG0_FIFODATA);
	}

	/*
	 * load command
	 */
	for (i = 0; i < SCpnt->cmd_len; i++)
	    outb (SCpnt->cmnd[i], info->scsi.io_port + REG0_FIFODATA);

	if (msglen == 1)
	    outb (CMD_SELECTATN, info->scsi.io_port + REG0_CMD);
	else
	    outb (CMD_SELECTATN3, info->scsi.io_port + REG0_CMD);
    } else
	outb (CMD_SELECTATNSTOP, info->scsi.io_port + REG0_CMD);

    printk (", data pointers [%p, %X]\n",
	    info->scsi.SCp.ptr, info->scsi.SCp.this_residual);

    /* should now get either DISCONNECT or (FUNCTION DONE with BUS SERVICE) intr */
}

/*
 * Function: void fas216_done (FAS216_Info *info, unsigned int result)
 * Purpose : complete processing for command
 * Params  : info   - interface that completed
 *	     result - driver byte of result
 */
static void fas216_done (FAS216_Info *info, unsigned int result)
{
    Scsi_Cmnd *SCpnt = info->SCpnt;

    if (info->scsi.aborting) {
	printk ("scsi%d.%c: uncaught abort - returning DID_ABORT\n",
		info->host->host_no, fas216_target (info));
	result = DID_ABORT;
	info->scsi.aborting = 0;
    }

    info->stats.fins += 1;

    if (SCpnt) {
    	info->scsi.phase = PHASE_IDLE;
	info->SCpnt = NULL;

	SCpnt->result = result << 16 | info->scsi.SCp.Message << 8 |
				info->scsi.SCp.Status;

	/*
	 * In theory, this should not happen, but just in case it does.
	 */
	if (info->scsi.SCp.ptr && result == DID_OK) {
	    switch (status_byte (SCpnt->result)) {
	    case CHECK_CONDITION:
	    case COMMAND_TERMINATED:
	    case BUSY:
	    case QUEUE_FULL:
	    case RESERVATION_CONFLICT:
		break;

	    default:
		printk (KERN_ERR "scsi%d.H: incomplete data transfer "
			"detected: result=%08X command=",
			info->host->host_no, SCpnt->result);
		print_command (SCpnt->cmnd);
	    }
	}
#if 0
	printk ("scsi%d.%c: scsi command (%p) complete, result=%08X\n",
		info->host->host_no, fas216_target (info),
		SCpnt, SCpnt->result);
#endif

	if (!SCpnt->scsi_done)
	    panic ("scsi%d.H: null scsi_done function in fas216_done", info->host->host_no);

	clear_bit (SCpnt->target * 8 + SCpnt->lun, info->busyluns);

	SCpnt->scsi_done (SCpnt);
    } else
	panic ("scsi%d.H: null command in fas216_done", info->host->host_no);

    if (info->scsi.irq != NO_IRQ)
	fas216_kick (info);
}

/*
 * Function: int fas216_queue_command (Scsi_Cmnd *SCpnt, void (*done)(Scsi_Cmnd *))
 * Purpose : queue a command for adapter to process.
 * Params  : SCpnt - Command to queue
 *	     done  - done function to call once command is complete
 * Returns : 0 - success, else error
 */
int fas216_queue_command (Scsi_Cmnd *SCpnt, void (*done)(Scsi_Cmnd *))
{
    FAS216_Info *info = (FAS216_Info *)SCpnt->host->hostdata;
#if 0
    printk ("scsi%d.%c: received queuable command (%p) %02X\n",
	    SCpnt->host->host_no, '0' + SCpnt->target,
	    SCpnt, SCpnt->cmnd[0]);
#endif
    SCpnt->scsi_done = done;
    SCpnt->host_scribble = NULL;
    SCpnt->result = 0;
    SCpnt->SCp.Message = 0;
    SCpnt->SCp.Status = 0;

    if (SCpnt->use_sg) {
	SCpnt->SCp.buffer = (struct scatterlist *) SCpnt->buffer;
	SCpnt->SCp.buffers_residual = SCpnt->use_sg - 1;
	SCpnt->SCp.ptr = (char *) SCpnt->SCp.buffer->address;
	SCpnt->SCp.this_residual = SCpnt->SCp.buffer->length;
    } else {
	SCpnt->SCp.buffer = NULL;
	SCpnt->SCp.buffers_residual = 0;
	SCpnt->SCp.ptr = (unsigned char *)SCpnt->request_buffer;
	SCpnt->SCp.this_residual = SCpnt->request_bufflen;
    }

    info->stats.queues += 1;
    SCpnt->tag = 0;

    if (info->scsi.irq != NO_IRQ) {
	unsigned long flags;

	/* add command into execute queue and let it complete under
	 * the drivers interrupts.
	 */
	if (!queue_add_cmd_ordered (&info->queues.issue, SCpnt)) {
	    SCpnt->result = DID_ERROR << 16;
	    done (SCpnt);
	}
	save_flags_cli (flags);
	if (!info->SCpnt || info->scsi.disconnectable)
	    fas216_kick (info);
	restore_flags (flags);
    } else {
	/* no interrupts to rely on - we'll have to handle the
	 * command ourselves.  For now, we give up.
	 */
	SCpnt->result = DID_ERROR << 16;
	done (SCpnt);
    }
    return 0;
}

/*
 * Function: void fas216_internal_done (Scsi_Cmnd *SCpnt)
 * Purpose : trigger restart of a waiting thread in fas216_command
 * Params  : SCpnt - Command to wake
 */
static void fas216_internal_done (Scsi_Cmnd *SCpnt)
{
    FAS216_Info *info = (FAS216_Info *)SCpnt->host->hostdata;

    info->internal_done = 1;
}

/*
 * Function: int fas216_command (Scsi_Cmnd *SCpnt)
 * Purpose : queue a command for adapter to process.
 * Params  : SCpnt - Command to queue
 * Returns : scsi result code
 */
int fas216_command (Scsi_Cmnd *SCpnt)
{
    FAS216_Info *info = (FAS216_Info *)SCpnt->host->hostdata;
    unsigned long flags;

    info->internal_done = 0;
    fas216_queue_command (SCpnt, fas216_internal_done);

    /*
     * This wastes time, since we can't return until the command is complete.
     * We can't seep either since we may get re-entered!
     * However, we must re-enable interrupts, or else we'll be waiting forever.
     */
    save_flags (flags);
    sti ();

    while (!info->internal_done)
	barrier ();

    restore_flags (flags);

    return SCpnt->result;
}

/*
 * Prototype: void fas216_reportstatus (Scsi_Cmnd **SCpntp1, Scsi_Cmnd **SCpntp2, int result)
 * Purpose  : pass a result to *SCpntp1, and check if *SCpntp1 = *SCpntp2
 * Params   : SCpntp1 - pointer to command to return
 *	      SCpntp2 - pointer to command to check
 *	      result  - result to pass back to mid-level done function
 * Returns  : *SCpntp2 = NULL if *SCpntp1 is the same command structure as *SCpntp2.
 */
static void fas216_reportstatus (Scsi_Cmnd **SCpntp1, Scsi_Cmnd **SCpntp2, int result)
{
    Scsi_Cmnd *SCpnt = *SCpntp1;

    if (SCpnt) {
	*SCpntp1 = NULL;

	SCpnt->result = result;
	SCpnt->scsi_done (SCpnt);
    }

    if (SCpnt == *SCpntp2)
	*SCpntp2 = NULL;
}

/*
 * Function: int fas216_abort (Scsi_Cmnd *SCpnt)
 * Purpose : abort a command if something horrible happens.
 * Params  : SCpnt - Command that is believed to be causing a problem.
 * Returns : one of SCSI_ABORT_ macros.
 */
int fas216_abort (Scsi_Cmnd *SCpnt)
{
    FAS216_Info *info = (FAS216_Info *)SCpnt->host->hostdata;
    int result = SCSI_ABORT_SNOOZE;

    info->stats.aborts += 1;

    printk (KERN_WARNING "fas216_abort:\n");

    if (queue_cmdonqueue (&info->queues.issue, SCpnt))
	printk ("scsi%d: command on issue queue\n", info->host->host_no);

    if (queue_cmdonqueue (&info->queues.disconnected, SCpnt))
	printk ("scsi%d: command on disconnected queue\n", info->host->host_no);

    if (info->SCpnt == SCpnt)
	printk ("scsi%d: command executing\n", info->host->host_no);

    if (info->origSCpnt == SCpnt) {
	info->origSCpnt = NULL;
	SCpnt->result = DID_ABORT << 16;
	SCpnt->scsi_done (SCpnt);
	printk ("scsi%d: command waiting for execution\n", info->host->host_no);
	result = SCSI_ABORT_SUCCESS;
    }

    return result;
}

/*
 * Function: int fas216_reset (Scsi_Cmnd *SCpnt, unsigned int reset_flags)
 * Purpose : resets the adapter if something horrible happens.
 * Params  : SCpnt - Command that is believed to be causing a problem.
 *	     reset_flags - flags indicating reset type that is believed to be required.
 * Returns : one of SCSI_RESET_ macros, or'd with the SCSI_RESET_*_RESET macros.
 */
int fas216_reset (Scsi_Cmnd *SCpnt, unsigned int reset_flags)
{
    FAS216_Info *info = (FAS216_Info *)SCpnt->host->hostdata;
    Scsi_Cmnd *SCptr;
    int result = 0;

    info->stats.resets += 1;

    printk (KERN_WARNING "fas126_reset\n");

    outb (info->scsi.cfg[3], info->scsi.io_port + REG0_CFG3);

    if (reset_flags & SCSI_RESET_SUGGEST_HOST_RESET) {
	outb (CMD_RESETCHIP, info->scsi.io_port + REG0_CMD);
	outb (CMD_NOP, info->scsi.io_port + REG0_CMD);
	result |= SCSI_RESET_HOST_RESET;
    }

    if (reset_flags & SCSI_RESET_SUGGEST_BUS_RESET) {
    	outb (CMD_RESETSCSI, info->scsi.io_port + REG0_CMD);
    	outb (CMD_NOP, info->scsi.io_port + REG0_CMD);
    	result |= SCSI_RESET_BUS_RESET;
    }

    if (!(reset_flags &
	  (SCSI_RESET_SUGGEST_BUS_RESET|SCSI_RESET_SUGGEST_HOST_RESET))) {
	outb (CMD_RESETCHIP, info->scsi.io_port + REG0_CMD);
	outb (CMD_NOP, info->scsi.io_port + REG0_CMD);
	outb (CMD_RESETSCSI, info->scsi.io_port + REG0_CMD);
	result |= SCSI_RESET_HOST_RESET | SCSI_RESET_BUS_RESET;
    }

    /*
     * Signal all commands in progress have been reset
     */
    fas216_reportstatus (&info->SCpnt, &SCpnt, DID_RESET);

    while ((SCptr = queue_remove (&info->queues.disconnected)) != NULL)
	fas216_reportstatus (&SCptr, &SCpnt, DID_RESET);

    if (SCpnt) {
	SCpnt->result = DID_RESET << 16;
	SCpnt->scsi_done (SCpnt);
    }

    return result | SCSI_RESET_SUCCESS;
}

/*
 * Function: int fas216_clockrate (unsigned int clock)
 * Purpose : calculate correct value to be written into clock conversion factor
 *	     register.
 * Params  : clock - clock speed in MHz
 * Returns : CCF_ value
 */
static int fas216_clockrate (int clock)
{
    if (clock <= 10 || clock > 40) {
	printk (KERN_CRIT "fas216: invalid clock rate: check your driver!\n");
	clock = -1;
    } else
	clock = ((clock - 1) / 5 + 1) & 7;

    return clock;
}

/*
 * Function: int fas216_init (struct Scsi_Host *instance)
 * Purpose : initialise FAS/NCR/AMD SCSI ic.
 * Params  : instance - a driver-specific filled-out structure
 * Returns : 0 on success
 */
int fas216_init (struct Scsi_Host *instance)
{
    FAS216_Info *info = (FAS216_Info *)instance->hostdata;
    unsigned long flags;
    int target_jiffies;

    info->host = instance;
    info->scsi.cfg[0] = instance->this_id;
    info->scsi.cfg[1] = CFG2_FENABLE | CFG2_SCSI2ENABLE;
    info->scsi.cfg[2] = CFG3_GROUP2 | 0x18;
    info->scsi.type = "unknown";
    info->scsi.phase = PHASE_IDLE;
    info->scsi.reconnected.target = 0;
    info->scsi.reconnected.lun = 0;
    info->scsi.reconnected.tag = 0;
    info->scsi.disconnectable = 0;
    info->SCpnt = NULL;

    memset (&info->stats, 0, sizeof (info->stats));

    msgqueue_initialise (&info->scsi.msgs);

    if (!queue_initialise (&info->queues.issue))
	return 1;

    if (!queue_initialise (&info->queues.disconnected)) {
	queue_free (&info->queues.issue);
	return 1;
    }

    outb (0, info->scsi.io_port + REG0_CFG3);
    outb (CFG2_SCSI2ENABLE, info->scsi.io_port + REG0_CFG2);

    if ((inb (info->scsi.io_port + REG0_CFG2) & (~0xe0)) != CFG2_SCSI2ENABLE) {
	info->scsi.type = "NCR53C90";
    } else {
	outb (0, info->scsi.io_port + REG0_CFG2);
	outb (0, info->scsi.io_port + REG0_CFG3);
	outb (5, info->scsi.io_port + REG0_CFG3);
	if (inb (info->scsi.io_port + REG0_CFG3) != 5) {
	    info->scsi.type = "NCR53C90A";
	} else {
	    outb (0, info->scsi.io_port + REG0_CFG3);
	    info->scsi.type = "NCR53C9x";
	}
    }


    outb (CFG3_IDENABLE, info->scsi.io_port + REG0_CFG3);
    outb (0, info->scsi.io_port + REG0_CFG3);

    outb (CMD_RESETCHIP, info->scsi.io_port + REG0_CMD);

    udelay (300);

    /* now for the real initialisation */
    outb (fas216_clockrate (info->ifcfg.clockrate), info->scsi.io_port + REG0_CLOCKCONV);
    outb (info->scsi.cfg[0], info->scsi.io_port + REG0_CFG1);
    outb (info->scsi.cfg[1], info->scsi.io_port + REG0_CFG2);
    outb (info->scsi.cfg[2], info->scsi.io_port + REG0_CFG3);
    outb (info->ifcfg.select_timeout, info->scsi.io_port + REG0_TIMEOUT);
    outb (0, info->scsi.io_port + REG0_SYNCOFFSET);
    outb (info->ifcfg.clockrate / 5, info->scsi.io_port + REG0_PERIOD);

    outb (info->scsi.cfg[0] | CFG1_DISABLERESETREP, info->scsi.io_port + REG0_CFG1);
    outb (CMD_RESETSCSI, info->scsi.io_port + REG0_CMD);

    /* scsi standard says 250ms */
    target_jiffies = jiffies + (25 * HZ) / 100;
    save_flags (flags);
    sti ();

    while (jiffies < target_jiffies) barrier ();

    restore_flags (flags);

    outb (info->scsi.cfg[0], info->scsi.io_port + REG0_CFG1);
    inb (info->scsi.io_port + REG0_ISR);

    return 0;
}

/*
 * Function: int fas216_release (struct Scsi_Host *instance)
 * Purpose : release all resources and put everything to bed for FAS/NCR/AMD SCSI ic.
 * Params  : instance - a driver-specific filled-out structure
 * Returns : 0 on success
 */
int fas216_release (struct Scsi_Host *instance)
{
    FAS216_Info *info = (FAS216_Info *)instance->hostdata;

    outb (CMD_RESETCHIP, info->scsi.io_port + REG0_CMD);
    queue_free (&info->queues.disconnected);
    queue_free (&info->queues.issue);

    return 0;
}
